#include <assert.h>
#include <errno.h>
#include <iostream.h>
#include <string.h>

#include "stdprj.h"
#include "machine.h"
#include "register.h"
#include "test1.h"
#include "test2.h"
#include "test3.h"
#include "test4.h"
#include "test5.h"
#include "test6.h"
#include "test7.h"
#include "test8.h"
#include "test9.h"
#include "test10.h"

// Cave!
//
// The following virtual functions from Machine must work correctly:
//
//        M.Reset()
//        M.Run()
//        M.IsHalted()
//        M.Step()
//
// The function M.Step() must be a mathematical mapping
//
//        M.Step(): M.Memory -> M.Memory
//
// In all of the functions above local static variables and
// access to global variables are not allowed. If you need
// any static or global storage, use M.Memory to implement
// your functions. But of course you can use local automatic
// variables.
//
// Only if the above conditions are met, InfLoop works correctly!
static int InfLoop(Machine & M, Register & R, Storage & S)
{
   // reset Machine M
   M.Reset();

   // reset Register R
   R.Reset();

   // Exit on error!
   assert(M.GetSize() == R.GetEntrySize());
   assert(M.GetSize() == S.GetSize());

   // start program on Machine M
   M.Run(S);
   // Exit on error!
   assert(M.IsRunning());

   // controlled run of program on Machine M
   //
   // M.Memory can take up only a finite number of states.
   // The Register R is big enough to handle all of these states.
   //
   // Hence: !!! THIS "while(...) {...}" IS NO INFINITE LOOP !!!
   while(M.IsRunning())
   {
      // register M.Memory
      R << M.Memory;

      // Exit with TRUE, if the state of M.Memory is already
      // registered (with other words, if the Machine M got
      // into an infinite loop).
      if (R.Found()) return(TRUE);

      // next step in the program of Machine M
      M.Step();
   }
   // Exit on error (which is actually impossible)!
   assert(M.IsHalted());
   // save last step into R
   R << M.Memory;

   // The Machine M has finished its program correctly.
   return(FALSE);
}

static void PrintResult(const char * NameMachine,
                        const char * NameProg,
                        int Result, Register & R)
{
   char *Path = (char *) NULL;
   char *P = (char *) NULL;
   int l;
   int x;

   // Exit on error!
   assert(NameMachine);
   l = strlen(NameMachine);
   assert(errno == EZERO);
   assert(l >= 1);
   assert(NameProg);

   if (Result)
   {
      if (strlen(NameProg) >= 1)
      {
         cout << "Infinite loop in the program "
              << NameProg
              << " on machine "
              << NameMachine
              << "!"
              << endl << endl;
         // Exit on error!
         assert(errno == EZERO);
      }
      else
      {
         cout << "Infinite loop in the program on machine "
              << NameMachine
              << "!"
              << endl << endl;
         // Exit on error!
         assert(errno == EZERO);
      }
   }
   else
   {
      if (strlen(NameProg) >= 1)
      {
         cout << "The program "
              << NameProg
              << " on machine "
              << NameMachine
              << " has terminated correctly "
              << endl
              << "without encountering any infinite loop."
              << endl << endl;
         // Exit on error!
         assert(errno == EZERO);
      }
      else
      {
         cout << "The program on machine "
              << NameMachine
              << " has terminated correctly "
              << endl
              << "without encountering any infinite loop."
              << endl << endl;
         // Exit on error!
         assert(errno == EZERO);
      }
   }

   // one char must have 8 bits
   assert(CHAR_BIT == 8);
   // test l for overflow on new char[l + 4 + 1]
   assert(4 < INT_MAX);
   assert(4 + 1 < INT_MAX - l);

   // allocate space from heap
   Path = new char[l + 4 + 1];
   // Exit on error!
   assert(Path);

   // generate filename
   P = strcpy(Path, NameMachine);
   // Exit on error!
   x = strcmp(Path, P);
   assert(errno == EZERO);
   assert(x == 0);
   assert(Path == P);

   if (strlen(NameProg) >= 1)
   {
      P = strcat(Path, NameProg);
      // Exit on error!
      x = strcmp(Path, P);
      assert(errno == EZERO);
      assert(x == 0);
      assert(Path == P);
   }

   P = strcat(Path, ".dmp");
   // Exit on error!
   x = strcmp(Path, P);
   assert(errno == EZERO);
   assert(x == 0);
   assert(Path == P);

   // Exit on error!
   x = strcmp(Path, P);
   assert(errno == EZERO);
   assert(x == 0);
   assert(Path == P);

   // dump R
   R.Dump(Path);

   // free allocated space
   delete [] Path;
}

extern int main()
{
   int Result;

   // test a program on Machine Test1 for infinite loop
   {
      Test1 T1;
      Register R1(T1.GetSize());
      Storage S1(T1.GetSize());

      Result = InfLoop(T1, R1, S1);
      PrintResult("Test1", "", Result, R1);
   }

   // test a program on Machine Test2 for infinite loop
   {
      Test2 T2;
      Register R2(T2.GetSize());
      Storage S2(T2.GetSize());

      Result = InfLoop(T2, R2, S2);
      PrintResult("Test2", "", Result, R2);
   }

   // test a program on Machine Test3 for infinite loop
   {
      Test3 T3;
      Register R3(T3.GetSize());
      Storage S3(T3.GetSize());

      Result = InfLoop(T3, R3, S3);
      PrintResult("Test3", "", Result, R3);
   }

   // test a program on Machine Test4 for infinite loop
   {
      Test4 T4;
      Register R4(T4.GetSize());
      Storage S4(T4.GetSize());

      Result = InfLoop(T4, R4, S4);
      PrintResult("Test4", "", Result, R4);
   }

   // test a program on Machine Test5 for infinite loop
   {
      Test5 T5;
      Register R5(T5.GetSize());
      Storage S5(T5.GetSize());

      Result = InfLoop(T5, R5, S5);
      PrintResult("Test5", "", Result, R5);
   }

   // test a program on Machine Test6 for infinite loop
   {
      Test6 T6;
      Register R6(T6.GetSize());
      Storage S6(T6.GetSize());

      Result = InfLoop(T6, R6, S6);
      PrintResult("Test6", "", Result, R6);
   }

   // test a program on Machine Test7 for infinite loop
   {
      Test7 T7;
      Register R7(T7.GetSize());
      Storage S7(T7.GetSize());

      Result = InfLoop(T7, R7, S7);
      PrintResult("Test7", "", Result, R7);
   }

   // test a program on Machine Test8 for infinite loop
   {
      Test8 T8;
      Register R8(T8.GetSize());
      Storage S8(T8.GetSize());

      Result = InfLoop(T8, R8, S8);
      PrintResult("Test8", "", Result, R8);
   }

   // test a program on Machine Test9 for infinite loop
   {
      Test9 T9;
      Register R9(T9.GetSize());
      Storage S9(T9.GetSize());

      Result = InfLoop(T9, R9, S9);
      PrintResult("Test9", "", Result, R9);
   }

   // test 2 programs on Machine Test10 for infinite loop
   {
      Test10 T10;
      Register R10(T10.GetSize());

      {
         Storage S10(T10.GetSize());

         (S10)[1] = 0x00;

         Result = InfLoop(T10, R10, S10);
         PrintResult("Test10", "1", Result, R10);
      }

      {
         Storage S10(T10.GetSize());

         (S10)[1] = 0x01;

         Result = InfLoop(T10, R10, S10);
         PrintResult("Test10", "2", Result, R10);
      }

   }

   return(0);
}

